// ---------------------------------------------------------------------------
// @file       larvorCommon.c
// @brief      fonction to read/write a buffer & CRC
// 
// @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @date       2009/09/25
// ---------------------------------------------------------------------------


#include "larvorCommon.h"
#include "uart.h"
#include "crc.h"
#include "larvorProtocol.h"
#include "time.h"

uint16_t addrBegin;
uint16_t addrEnd;
uint8_t  pageSize;

// ----------------------------------------------------------------------------
// Function which read 2 bytes on the serial port. 
// The first byte is copied in 'data'
// The second byte must must be the CRC8 of 'data'. Return 1 if thisis the case
// ----------------------------------------------------------------------------
unsigned char 
larvorGetDataAndCrc(unsigned char* data)
{
    *data = uartRead8();
    unsigned char crc = uartRead8();
    unsigned char crcData = CRC8(*data, 0); 
    return (crcData == crc)?1:0;
}

// ----------------------------------------------------------------------------
// Read 2 16 bits addresses + the 8 bits page size + 8 bits CRC for these data
// ----------------------------------------------------------------------------
unsigned char 
larvorGetAddrs()
{
    addrBegin = uartRead16();
    addrEnd   = uartRead16();
    pageSize  = uartRead8();
    unsigned char crc = uartRead8();
    
    // check CRC
    unsigned char crcData = 0;
    crcData = CRC16(addrBegin, crcData); 
    crcData = CRC16(addrEnd, crcData); 
    crcData = CRC8(pageSize, crcData); 
    
    if (crcData == crc) {
        uartWrite(PROG_CRC_OK);
    } else {
        uartWrite(PROG_CRC_ERROR);
    }
    return (crcData == crc)?1:0;
}



#define UART_TIMEOUT 1000 // 1000 ms = 1s

// ----------------------------------------------------------------------------
// AVR read data from the serial port and check the CRC
// ----------------------------------------------------------------------------
static unsigned char rxBuffer[140]; // max 128 + 1 status byte + 1 CRC

static unsigned char 
larvorGetWriteBuffer(unsigned char** writeBuffer,
                          unsigned char bufferSize)
{
    // copy the data from the serial port fifo
    timeReset();
    while(1) {
        if (uartReadNBytes(rxBuffer, bufferSize + 2) != 0) // data received
            break;
        if (timeGet() > UART_TIMEOUT) {// timeout
            uartWrite(PROG_TIMEOUT);
            return PROG_CANCEL;
        }
    }
    // check the COM status byte
    if (rxBuffer[0] == PROG_CANCEL) { // communication stopped by the PC
        uartWrite(PROG_CANCEL);
        return PROG_CANCEL;
    }

    // set the data pointer
    *writeBuffer = rxBuffer + 1;

    // compute the CRC
    unsigned char i = 0;
    unsigned char crc=0;
    while(i < bufferSize) {
        crc = CRC8((*writeBuffer)[i++], crc);
    }

    // check the CRC value to be sure that there was no error on the serial port
    uint8_t crcExpected = rxBuffer[1 + bufferSize];
    if (crc != crcExpected) {
        uartWrite(PROG_CRC_ERROR);
        return PROG_CRC_ERROR;
    } else {
        uartWrite(PROG_CRC_OK);
        return PROG_CRC_OK;
    }
}

// ----------------------------------------------------------------------------
// resynchronize the UART connection after a CRC error (probably a missing character)
// ----------------------------------------------------------------------------
static void
larvorUartResync()
{
    timeReset();
    while(uartRead(UART_DONT_WAIT_RECEIVE) != PROG_SYNC1 && timeGet() < UART_TIMEOUT);
    while(uartRead(UART_DONT_WAIT_RECEIVE) != PROG_SYNC2 && timeGet() < UART_TIMEOUT);
}

// ----------------------------------------------------------------------------
// 
// ----------------------------------------------------------------------------
uint8_t
larvorGetWriteBufferData(uint8_t** buffer, 
                         uint8_t size)
{
    uint8_t rcvStatus = PROG_CRC_ERROR;
    do {
        rcvStatus = larvorGetWriteBuffer(buffer, size);
        if (rcvStatus == PROG_CRC_ERROR) {
            larvorUartResync();
        }
    } while (rcvStatus == PROG_CRC_ERROR);
    return rcvStatus;
}


