/**
 * @file       led.c
 * @brief      led driver
 * 
 * @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
 * @date       2004/07/25
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "led.h"
#include "util.h"

#define LED_IO_PORT   PORTC
#define LED_IO_DIR    DDRC
#define LED_FIRST_PIN 0

/* ------------------------------------------------------------------------ */
/* static                                                                   */
/* ------------------------------------------------------------------------ */
static struct {
    uint8_t   ledMode_[LED_NBR];
    uint8_t   ledBit_;
    uint8_t   ledTimer_;
} LED;

/* ------------------------------------------------------------------------ */
/* ISR(SIG_OVERFLOW0)                                                       */
/* ------------------------------------------------------------------------ */
/* Interrupt when Timer 0 reached overflow                                  */
/* Turn ON/OFF the leds every 150ms depending on their mode                 */
/* ------------------------------------------------------------------------ */
ISR(SIG_OVERFLOW0)  
{
    // every 8.192 ms
	++LED.ledTimer_;
    if (LED.ledTimer_ < 18) {
        return;
    }
    // every 150 ms
    LED.ledTimer_ = 0;

    ++LED.ledBit_;
    if (LED.ledBit_ >= 8 ) {
        LED.ledBit_ = 0;
    }
    uint8_t i;
    for(i=0; i < LED_NBR; i++) {
        if (LED.ledMode_[i] & (1 << LED.ledBit_)) {
            sbi(LED_IO_PORT, LED_FIRST_PIN + i);
        } else {
            cbi(LED_IO_PORT, LED_FIRST_PIN + i);
        }
    }
}

/* ------------------------------------------------------------------------ */
/* ledInit                                                                */
/* ------------------------------------------------------------------------ */
void ledInit()
{
    /* Define IO ports as output, lower state */
    /* LEDS */
	LED.ledTimer_ = 0;
    LED.ledBit_   = 0;
    uint8_t i;
    for(i=0; i < LED_NBR; i++) {
        sbi(LED_IO_DIR, LED_FIRST_PIN + i);
        ledSetMode(i, LED_MODE_OFF);
    }

    /* Set timer 0 speed : CLK/256 = 0.032ms/tick */
    TCCR0 = 0x04;
    /* Set timer 0 value, when reach overflow, INTERRUPT(SIG_OVERFLOW0) is run */
    TCNT0 = 0;
    /* Enable timer 0 interrupt */
    sbi(TIMSK, TOIE0);
}

/* ------------------------------------------------------------------------ */
/* ledSetMode                                                               */
/* ------------------------------------------------------------------------ */
void ledSetMode(uint8_t ledId,
                uint8_t mode)
{
    LED.ledMode_[ledId] = mode;
}
