/**
 * @file       main.c
 * @brief      Interface between serial port and JTAG/SPI  
 * 
 * @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
 * @date       2009/09/25
 */

#include "uart.h"
#include "jtag.h"
#include "avrJtag.h"
#include "spi.h"
#include "crc.h"
#include "util.h"
#include "led.h"
#include "time.h"
#include "larvorProtocol.h"
#include "larvorCommon.h"

#include <avr/interrupt.h>

// ----------------------------------------------------------------------------
// Function to communicate with the programmer board on serial port
// ----------------------------------------------------------------------------
void
cmdInterpreter()
{
    unsigned char req = uartRead(UART_WAIT_RECEIVE);
	unsigned char data1;
    switch(req) {
    // -----------------------------------------------------------------------
	// BASIC COMMANDS
	// -----------------------------------------------------------------------
	case PROG_PING:
        uartWrite(PROG_PING);
        return;
    case PROG_SET_LED_1:
        ledSetMode(0, uartRead8());
        return;
    case PROG_SET_LED_2:
        ledSetMode(1, uartRead8());
        return;

    // -----------------------------------------------------------------------
	// JTAG READ
	// -----------------------------------------------------------------------
	case PROG_JTAG_GET_DEVICE_ID:
		uartWrite32(avrJtagGetDeviceId());
		break;
	case PROG_JTAG_READ_SIGNATURE:
		uartWrite32(avrJtagReadSignature());
		break;
	case PROG_JTAG_READ_CALIBRATION:
        uartWrite8(avrJtagReadCalibration());
        break;
    case PROG_JTAG_READ_FUSE_EXT:
        uartWrite8(avrJtagReadFuseExt());
        break;
    case PROG_JTAG_READ_FUSE_HIGH:
        uartWrite8(avrJtagReadFuseHigh());
        break;
    case PROG_JTAG_READ_FUSE_LOW:
        uartWrite8(avrJtagReadFuseLow());
        break;
    case PROG_JTAG_READ_LOCK:
        uartWrite8(avrJtagReadLock());
        break;

	// -----------------------------------------------------------------------
	// JTAG WRITE
	// -----------------------------------------------------------------------
	case PROG_JTAG_WRITE_FUSE_EXT:
        if (larvorGetDataAndCrc(&data1)) {
            avrJtagWriteFuseExt(data1);
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
        }
        return;
    case PROG_JTAG_WRITE_FUSE_HIGH:
        if (larvorGetDataAndCrc(&data1)) {
            avrJtagWriteFuseHigh(data1);
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
        }
        return;
    case PROG_JTAG_WRITE_FUSE_LOW:
        if (larvorGetDataAndCrc(&data1)) {
            avrJtagWriteFuseLow(data1);
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
        }
       return;
    case PROG_JTAG_WRITE_LOCK:
        uartWrite(PROG_ERROR);
        return; // NOT Activated to avoid modifying locks accidentally
        if (larvorGetDataAndCrc(&data1)) {
            avrJtagWriteLock(data1); 
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
            
        }
        return;

	// -----------------------------------------------------------------------
	// JTAG EEPROM + FLASH PROGRAMMING
	// -----------------------------------------------------------------------
    case PROG_JTAG_CHIP_ERASE:
        avrJtagChipErase();
        break;

    case PROG_JTAG_READ_EEPROM:
        avrJtagReadEEprom();
        return;
    case PROG_JTAG_WRITE_EEPROM:
        avrJtagWriteEEprom();
        return;
    case PROG_JTAG_READ_FLASH:
        avrJtagReadFlash();
        return;
    case PROG_JTAG_WRITE_FLASH:
        avrJtagWriteFlash();
        return;

	// -----------------------------------------------------------------------
	// SPI READ
	// -----------------------------------------------------------------------
	case PROG_SPI_GET_DEVICE_ID:
		uartWrite32(spiGetDeviceId());
		break;
	case PROG_SPI_READ_SIGNATURE:
		uartWrite32(spiReadSignature());
		break;
	case PROG_SPI_READ_CALIBRATION:
        uartWrite8(spiReadCalibration());
        break;
    case PROG_SPI_READ_FUSE_EXT:
        uartWrite8(spiReadFuseExt());
        break;
    case PROG_SPI_READ_FUSE_HIGH:
        uartWrite8(spiReadFuseHigh());
        break;
    case PROG_SPI_READ_FUSE_LOW:
        uartWrite8(spiReadFuseLow());
        break;
    case PROG_SPI_READ_LOCK:
        uartWrite8(spiReadLock());
        break;

	// -----------------------------------------------------------------------
	// SPI WRITE
	// -----------------------------------------------------------------------
	case PROG_SPI_WRITE_FUSE_EXT:
        if (larvorGetDataAndCrc(&data1)) {
            spiWriteFuseExt(data1);
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
        }
        return;
    case PROG_SPI_WRITE_FUSE_HIGH:
        if (larvorGetDataAndCrc(&data1)) {
            spiWriteFuseHigh(data1);
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
        }
        return;
    case PROG_SPI_WRITE_FUSE_LOW:
        if (larvorGetDataAndCrc(&data1)) {
            spiWriteFuseLow(data1);
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
        }
       return;
    case PROG_SPI_WRITE_LOCK:
        uartWrite(PROG_ERROR);
        return; // NOT Activated to avoid modifying locks accidentally
        if (larvorGetDataAndCrc(&data1)) {
            spiWriteLock(data1); 
			uartWrite(PROG_OK);
        } else {
            uartWrite(PROG_ERROR);
            
        }
        return;

	// -----------------------------------------------------------------------
	// SPI EEPROM + FLASH PROGRAMMING
	// -----------------------------------------------------------------------
    case PROG_SPI_CHIP_ERASE:
        spiChipErase();
        break;

    case PROG_SPI_READ_EEPROM:
        spiReadEEprom();
        return;
    case PROG_SPI_WRITE_EEPROM:
        spiWriteEEprom();
        return;
    case PROG_SPI_READ_FLASH:
        spiReadFlash();
        return;
    case PROG_SPI_WRITE_FLASH:
        spiWriteFlash();
        return;

	
	
	// -----------------------------------------------------------------------
	// Default
	// -----------------------------------------------------------------------
    default:
        uartWrite(PROG_ERROR);
        return;
    }
    uartWrite(PROG_OK);
}

// ----------------------------------------------------------------------------
// Main function
// ----------------------------------------------------------------------------

int main(void)
{
    // init each module
    uartInit(UART_BAUD_RATE_38400);
    jtagInit();
	ledInit();
	timeInit();
	spiInit();
    //resetInit();
    sei(); // enable interrupts

	while(true) {
        cmdInterpreter();
    }
}
