/**
 * @file       spi.c
 * @brief      SPI communication for AVR 
 * 
 * @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
 * @date       2009/09/25
 */
#include <avr/io.h>
#include <util/delay.h>

#include "util.h"
#include "spi.h"
#include "larvorProtocol.h"
#include "larvorCommon.h"
#include "crc.h"
#include "uart.h"

// ----------------------------------------------------------------------------
// Function to read/write on SPI port of the programmer board
// ----------------------------------------------------------------------------

// pin configuration
#define SPI_DDR         DDRB
#define SPI_D0_PORT     PORTB
#define SPI_DI_PORT     PINB
#define SPI_CLK_PORT    PORTB
#define SPI_RESET_PORT  PORTB
#define SPI_DO_PIN      5
#define SPI_DI_PIN      6
#define SPI_CLK_PIN     7
#define SPI_RESET_PIN   4  

// ----------------------------------------------------------------------------
// sleep for a specifed amout of time
// ----------------------------------------------------------------------------
static void 
spiSleep(unsigned char timeMS)
{
	double t = timeMS * 1.0;
	_delay_ms(t);
}

uint16_t _currentDeviceId=0;
static uint16_t
spiGetCurrentDeviceId()
{
	return _currentDeviceId;
}

// ---------------------------------------------------------------------------
// Set the Jtag Reset pin state
// ---------------------------------------------------------------------------
static void 
spiReset(unsigned char reset)
{
    if (reset) {
        cbi(SPI_RESET_PORT, SPI_RESET_PIN);
    } else {
        sbi(SPI_RESET_PORT, SPI_RESET_PIN);
    }
}

// ---------------------------------------------------------------------------
// Send a byte on the SPI and return the answer
// ---------------------------------------------------------------------------
static unsigned char 
spiWrite(unsigned char data) 
{  
    // send character over SPI
    SPDR = data;
    // wait the answer and return it
    while(!(SPSR & (1<<SPIF)));

    return SPDR;
}

// ----------------------------------------------------------------------------
// AVR spi Enter programming mode
// ----------------------------------------------------------------------------
static unsigned char
spiProgrammingEnable()
{
    // set the reset pin to low
    spiReset(1);
    spiSleep(20); //sleep20 ms

    unsigned char deviceDetected=0;
    unsigned char retry=0;
    while(retry++ < 32 && !deviceDetected ) {
        spiWrite(0xAC);
        spiWrite(0x53);
        unsigned char echo = spiWrite(0xFF);
        spiWrite(0xFF);
        deviceDetected = (echo == 0x53);
    }
    return deviceDetected;
}

// ----------------------------------------------------------------------------
// AVR spi leave programming mode
// ----------------------------------------------------------------------------
static void
spiLeavingProgramming()
{
    // set the reset pin to high
    spiReset(0);
}

// ---------------------------------------------------------------------------
// Initialize the SPI port
// ---------------------------------------------------------------------------
void
spiInit()
{
    // init SPI port
    cbi(SPI_DDR, SPI_DI_PIN);     // set port B SPI data (MISO) input to input
    sbi(SPI_DDR, SPI_CLK_PIN);	  // set port B SPI clock to output
    sbi(SPI_DDR, SPI_DO_PIN);	  // set port B SPI data out (MOSI)to output
	sbi(SPI_DDR, SPI_RESET_PIN);  // set port B SPI reset pin to output

    SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0);
}

// ----------------------------------------------------------------------------
// AVR spi Reading the eeprom from addrBegin to addrEnd (addrEnd is included)
// ----------------------------------------------------------------------------
void
spiChipErase()
{
    if (spiProgrammingEnable()) {
        spiWrite(0xAC);
        spiWrite(0x9F);
        spiWrite(0xFF);
        spiWrite(0xFF);

        // sleep Twd_erase ( 8ms at 5V, 18ms a 3.2V)
        // sleep 20 ms for at90s2313
		spiSleep(20);
    }
    spiLeavingProgramming();
}


// ----------------------------------------------------------------------------
// AVR spi Reading signature
// ----------------------------------------------------------------------------
uint32_t
spiReadSignature()
{
    uint32_t signature=0;

    if (spiProgrammingEnable()) {
        spiWrite(0x30);
        spiWrite(0xFF);
        spiWrite(0x00);
		signature = ((uint32_t)spiWrite(0xFF)) << 16;

        spiWrite(0x30);
        spiWrite(0xFF);
        spiWrite(0x01);
        signature += ((uint32_t)spiWrite(0xFF)) << 8;

        spiWrite(0x30);
        spiWrite(0xFF);
        spiWrite(0x02);
        signature += ((uint32_t)spiWrite(0xFF)) << 0;
    }
    spiLeavingProgramming();

	_currentDeviceId = signature & 0xFFFF;
    return signature;
}

// ----------------------------------------------------------------------------
// AVR spi return the device Id as avrJtag getDeviceId
// ----------------------------------------------------------------------------
uint32_t
spiGetDeviceId()
{
	uint32_t signature = spiReadSignature();
	uint32_t deviceId = signature & 0xFFFF;
	deviceId = deviceId << 12;
	deviceId += 0x3F;
	return deviceId;
}

// ----------------------------------------------------------------------------
// AVR spi Reading calibration byte
// ----------------------------------------------------------------------------
uint8_t
spiReadCalibration()
{
	uint8_t calibration=0;
    if (spiProgrammingEnable()) {
		// TODO
        switch(spiGetCurrentDeviceId()) {
        case 0x9306: // 8515
            spiWrite(0x38);
            spiWrite(0x3F);
            spiWrite(0x00);
            calibration = spiWrite(0xFF);
            break;
        case 0x9508:
        case 0x940A:
            spiWrite(0x30);
            spiWrite(0xFF);
            spiWrite(0xFC);
            calibration = spiWrite(0xFF);
            break;
        default:
            break;
        }
    }

    spiLeavingProgramming();
    return calibration;
}

// ----------------------------------------------------------------------------
// AVR spi Reading the Fuse Extended byte
// ----------------------------------------------------------------------------
uint8_t
spiReadFuseExt()
{
    return 0xFF;
}

// ----------------------------------------------------------------------------
// AVR spi Reading the Fuse High byte
// ----------------------------------------------------------------------------
uint8_t
spiReadFuseHigh()
{
    uint8_t fuseHigh=0;
    if (spiProgrammingEnable()) {
        spiWrite(0x58);   
        spiWrite(0x08);
        spiWrite(0xFF);
        fuseHigh = spiWrite(0xFF);
    }
    spiLeavingProgramming();
    return fuseHigh;
}

// ----------------------------------------------------------------------------
// AVR spi Reading the Fuse Low byte
// ----------------------------------------------------------------------------
uint8_t 
spiReadFuseLow()
{
    uint8_t fuseLow=0;
    if (spiProgrammingEnable()) {
        spiWrite(0x50);   
        spiWrite(0x00);
        spiWrite(0xFF);
        fuseLow = spiWrite(0xFF);
    }
    spiLeavingProgramming();
    return fuseLow;
}

// ----------------------------------------------------------------------------
// AVR spi Reading the Lock byte
// ----------------------------------------------------------------------------
uint8_t 
spiReadLock()
{
    uint8_t lock=0;
    if (spiProgrammingEnable()) {
        spiWrite(0x58);   
        spiWrite(0x00);
        spiWrite(0xFF);
        lock = spiWrite(0xFF);
        
    }
    spiLeavingProgramming();
    return lock;
}

// ----------------------------------------------------------------------------
// AVR spi Writing the Fuse Extended byte
// ----------------------------------------------------------------------------
void
spiWriteFuseExt(uint8_t data)
{
    // TODO
}

// ----------------------------------------------------------------------------
// AVR spi Writing the Fuse High byte
// ----------------------------------------------------------------------------
void
spiWriteFuseHigh(uint8_t data)
{
    if (spiProgrammingEnable()) { 
        spiWrite(0xAC);   
        spiWrite(0xA8);
        spiWrite(0xFF);
        spiWrite(data);
    }
    spiLeavingProgramming();
}

// ----------------------------------------------------------------------------
// AVR spi Writing the Fuse Low bytes
// ----------------------------------------------------------------------------
void 
spiWriteFuseLow(uint8_t data)
{
    if (spiProgrammingEnable()) {
		// TODO
        switch(spiGetCurrentDeviceId()) {
        case 0x9306:
        case 0x9508:
        case 0x940A:
        case 0x960A:
            spiWrite(0xAC);   
            spiWrite(0xA0);
            spiWrite(0xFF);
            spiWrite(data);
            break;
        case 0x9203:
            spiWrite(0xAC);   
            spiWrite(0xA0 + (data&0x1F) );
            spiWrite(0xFF);
            spiWrite(0xFF);
            break;
        default:
            break;
        }
    }
}
// ----------------------------------------------------------------------------
// AVR jtag Writing the Lock bytes
// ----------------------------------------------------------------------------
void
spiWriteLock(uint8_t data)
{
    if (spiProgrammingEnable()) {
        spiWrite(0xAC);   
        spiWrite(0xFF);
        spiWrite(0xFF);
		spiWrite(data | 0xC0);
    }
    spiLeavingProgramming();
}

// ----------------------------------------------------------------------------
// AVR jtag Reading a eeprom block
// ----------------------------------------------------------------------------
static uint8_t 
spiEEpromRead(uint16_t begin,
              uint16_t size,
              uint8_t* compareBuffer,
              uint8_t* crc)
{
    uint16_t i=0, index=0;
    for(index=0; index < size; ++index) {
        uint16_t addr= begin + index;
        uint8_t addrL = ((addr) & 0xFF);
        uint8_t addrH = ((addr >> 8) & 0xFF);
		
		spiWrite(0xA0);
		spiWrite(addrH & 0x3F);
		spiWrite(addrL);
		uint8_t data, lastData = spiWrite(0xFF);

        while(1) {
            spiWrite(0xA0);
			spiWrite(addrH & 0x3F);
			spiWrite(addrL);
			data = spiWrite(0xFF);
            if (data == lastData) {
                break;
            } else {
                lastData = data;
            }
        }

        if (compareBuffer != 0) {
            // compare data with the buffer
            if (compareBuffer[i] != data) {
                return PROG_ERROR;
            }
            i++;
        } else {
            // send data on the serial port
            uartWrite8(data);
            // CRC
            *crc = CRC8(data, *crc);
        }
    } 
    return PROG_OK;
}
#define spiCompareEEpromBlock(addr, size, buffer)  spiEEpromRead(addr, size, buffer, 0)
#define spiSendEEpromBlock(addr, size, crc)        spiEEpromRead(addr, size, 0, crc)

// ----------------------------------------------------------------------------
// AVR jtag Writing the EEprom page
// ----------------------------------------------------------------------------
static void 
spiWriteEEpromPage(uint16_t addr,
                   uint8_t  pageSize,
                   uint8_t* buffer)
{
    uint8_t addrH = ((addr >> 8) & 0xFF);
    // save data on the jtag port
    uint8_t indexInPage = 0;
    for(indexInPage=0; indexInPage < pageSize; ++indexInPage) {
        uint8_t addrL = ((addr+indexInPage) & 0xFF);
		spiWrite(0xC0);
		spiWrite(addrH & 0x3F);
		spiWrite(addrL);
		spiWrite(buffer[indexInPage]);
		// TODO: sleep?
		spiSleep(9);
    }
}

// ----------------------------------------------------------------------------
// AVR spi Reading the eeprom from addrBegin to addrEnd (addrEnd is included)
// ----------------------------------------------------------------------------
void
spiReadEEprom()
{
	if (!larvorGetAddrs()) {
        return;
    }

    if (spiProgrammingEnable()) {
		uint16_t addr;
		for (addr = addrBegin; addr <= addrEnd;) {
			uint8_t crc=0;
			// read a block, send it on the UART and compute the CRC
			spiSendEEpromBlock(addr, PROG_CRC_CHECK_SIZE, &crc);
			// send the CRC
			uartWrite8(crc);
			// check the response from PC
			uint8_t answer = uartRead8();
			if (answer == PROG_CANCEL) {
				break;
			} else if (answer == PROG_CRC_OK) {
				// move to the next block
				addr = addr + PROG_CRC_CHECK_SIZE;
			} else { // PROG_CRC_ERROR
				// send the same block again
			}
		} 
	}
    spiLeavingProgramming();
    uartWrite(PROG_OK);
}

// ----------------------------------------------------------------------------
// AVR spi Writing the eeprom 
// ----------------------------------------------------------------------------
void
spiWriteEEprom()
{
	if (!larvorGetAddrs()) {
        return;
    }
    
    uint8_t status = PROG_OK;
    uint16_t addr=0;
	if (spiProgrammingEnable()) {
		for(addr=addrBegin; addr <= addrEnd; addr += pageSize) {
			// get data from the serial port
			uint8_t* buffer=0;
			if (larvorGetWriteBufferData(&buffer, pageSize) != PROG_CRC_OK) {
				break;
			}
			uint8_t retry;
			for(retry = 0; retry < 5; retry++) {
				// write data into the flash
				spiWriteEEpromPage(addr, pageSize, buffer);
				// check if data into the flash map the buffer
				if (spiCompareEEpromBlock(addr, pageSize, buffer) == PROG_OK)
					break;
			}
			if (retry >= 5) {
				status = PROG_ERROR;
				break;
			}
		}
	}

    spiLeavingProgramming();
    uartWrite(status);
}

// ----------------------------------------------------------------------------
// AVR spi Reading a flash block
// ----------------------------------------------------------------------------
static uint8_t 
spiFlashRead(uint16_t begin,
             uint16_t size,
             uint8_t* compareBuffer,
             uint8_t* crc)
{
    uint16_t i=0, index=0;
    for(index=0; index < size; ++index) {
        uint16_t addr= begin + index;
        uint8_t addrH = ((addr >> 8) & 0xFF);
        uint8_t addrL = (addr & 0xFF);
		
		// read data at least 2 times to be sure that we have read the correct value (lots of error append during the read)
        uint8_t dataH, dataL, lastDataH, lastDataL;
        
		spiWrite(0x20);
		spiWrite(addrH);
		spiWrite(addrL);
		lastDataL = spiWrite(0xFF);
		
		spiWrite(0x28);
		spiWrite(addrH);
		spiWrite(addrL);
		lastDataH = spiWrite(0xFF);
        while(1) {
            spiWrite(0x20);
			spiWrite(addrH);
			spiWrite(addrL);
			dataL = spiWrite(0xFF);
		
			spiWrite(0x28);
			spiWrite(addrH);
			spiWrite(addrL);
			dataH = spiWrite(0xFF);
            if ((dataH == lastDataH) && (dataL == lastDataL)) {
                break;
            } else {
                lastDataH = dataH;
                lastDataL = dataL;
            }
        }

        if (compareBuffer != 0) {
            // compare data with the buffer
            if ((compareBuffer[i] != dataL)
                || (compareBuffer[i+1] != dataH))
                {
                return PROG_ERROR;
            }
            i+=2;
        } else {
            // send data on the serial port
            uartWrite8(dataL);
            uartWrite8(dataH);
            // CRC
            *crc = CRC8(dataL, *crc);
            *crc = CRC8(dataH, *crc);
        }
    } 
    return PROG_OK;
}
#define spiCompareFlashBlock(addr, size, buffer)  spiFlashRead(addr, size, buffer, 0)
#define spiSendFlashBlock(addr, size, crc)        spiFlashRead(addr, size, 0, crc)
// ----------------------------------------------------------------------------
// AVR spi Writing the flash page
// ----------------------------------------------------------------------------
static void 
spiWriteFlashPage(uint16_t addr,
                  uint8_t  pageSize,
                  uint8_t* buffer)
{
    uint8_t addrH = ((addr >> 8) & 0xFF);
	uint8_t addrL=0;

    // save data on the jtag port
    uint8_t indexInPage = 0;
    for(indexInPage=0; indexInPage < pageSize; ++indexInPage) {
        uint8_t addrL = ((addr+indexInPage) & 0xFF);

        uint8_t dataL = buffer[indexInPage*2];
        uint8_t dataH = buffer[indexInPage*2+1];
		
		spiWrite(0x40);
		spiWrite(addrH);
		spiWrite(addrL);
		spiWrite(dataL);

		spiSleep(9); // 9ms at 3.5V 4ms a 5V
		// sleep ?
		
		spiWrite(0x48);
		spiWrite(addrH);
		spiWrite(addrL);
		spiWrite(dataH);
		// sleep ?
		spiSleep(9);
    }
	
	if (pageSize > 1) {
		// write program memory page
		spiWrite(0x4C);
		spiWrite(addrH);
		spiWrite(addrL);
		spiWrite(0xFF);
	}
}

// ----------------------------------------------------------------------------
// AVR spi Reading the flash
// ----------------------------------------------------------------------------
void
spiReadFlash()
{
    if (!larvorGetAddrs()) {
        return;
    }

	if (spiProgrammingEnable()) {
		uint16_t addr;
		for (addr = addrBegin; addr <= addrEnd;) {
			uint8_t crc=0;
			// read a block, send it on the UART and compute the CRC
			spiSendFlashBlock(addr, PROG_CRC_CHECK_SIZE, &crc);
			// send the CRC
			uartWrite8(crc);
			// check the response from PC
			uint8_t answer = uartRead8();
			if (answer == PROG_CANCEL) {
				break;
			} else if (answer == PROG_CRC_OK) {
				// move to the next block
				addr = addr + PROG_CRC_CHECK_SIZE;
			} else { // PROG_CRC_ERROR
				// send the same block again
			}
		} 
	}

    spiLeavingProgramming();
    uartWrite(PROG_OK);
}

// ----------------------------------------------------------------------------
// AVR spi Writing the flash
// ----------------------------------------------------------------------------
void 
spiWriteFlash()
{
    if (!larvorGetAddrs()) {
        return;
    }

    uint8_t status = PROG_OK;
    uint16_t addr=0;
	if (spiProgrammingEnable()) {
		for(addr=addrBegin; addr <= addrEnd; addr += pageSize) {
			// get data from the serial port
			uint8_t* buffer=0;
			if (larvorGetWriteBufferData(&buffer, pageSize*2) != PROG_CRC_OK) {
				break;
			}
			uint8_t retry;
			for(retry = 0; retry < 5; retry++) {
				// write data into the flash
				spiWriteFlashPage(addr, pageSize, buffer);
				// check if data into the flash map the buffer
				if (spiCompareFlashBlock(addr, pageSize, buffer) == PROG_OK)
					break;
			}
			if (retry >= 5) {
				status = PROG_ERROR;
				break;
			}
		}
	}

    spiLeavingProgramming();
    uartWrite(status);
}

