/**
 * @file       uart.c
 * @brief      serial port communication for AVR (ATmega32)
 * 
 * @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
 * @date       2004/07/25
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include "uart.h"

// ----------------------------------------------------------------------------
// nop
// ----------------------------------------------------------------------------
static void nop() 
{ 
	__asm__ __volatile__ ("nop"); 
}



#ifdef UART_RX_FIFO
static struct UartRxBuffer
{
	volatile unsigned char data[UART_RX_BUFFER_SIZE];
	         unsigned char readIndex;
	volatile unsigned char writeIndex;
	volatile unsigned char dataLength;
} rxBuffer; 

/* ------------------------------------------------------------------------*/
/* uartPopRxBuffer                                                         */
/* ------------------------------------------------------------------------*/
unsigned char uartReadNBytes(unsigned char* buffer,
						     unsigned char length)
{
	if ((rxBuffer.dataLength > 0) && (rxBuffer.dataLength >= length)) {
		unsigned char i=0;
		for(i=0; i < length; ++i) {
			buffer[i] = rxBuffer.data[rxBuffer.readIndex];
			if (++rxBuffer.readIndex >= UART_RX_BUFFER_SIZE) {
				rxBuffer.readIndex = 0;
			}
			--rxBuffer.dataLength;
		}
		return 1;
	}
	return 0;
}

/* ------------------------------------------------------------------------ */
/* ISR(SIG_UART_RECV)                                                       */
/* ------------------------------------------------------------------------ */
/* Interrupt when data are received                                          */
/* ------------------------------------------------------------------------ */
ISR(SIG_UART_RECV)  
{
	unsigned char data = UDR;
	if (rxBuffer.dataLength < UART_RX_BUFFER_SIZE) {
		rxBuffer.data[rxBuffer.writeIndex] = data;
		if (++rxBuffer.writeIndex >= UART_RX_BUFFER_SIZE) {
			rxBuffer.writeIndex = 0;
		}
		++rxBuffer.dataLength;
	}
}

/* ------------------------------------------------------------------------*/
/* uartRead                                                                */
/* ------------------------------------------------------------------------*/
unsigned char uartRead(unsigned char blocking)
{
    if (blocking == UART_WAIT_RECEIVE) {
        /* wait for data register to be full */
        while ( rxBuffer.dataLength == 0) 
			nop();
    }
    /* data register not empty, return the data */
	unsigned char data;
	if (uartReadNBytes(&data, 1)) {
		return data;
	} else {
		return UART_NO_RCV;
	}
}

#else // UART_RX_FIFO
/* ------------------------------------------------------------------------*/
/* uartRead                                                                */
/* ------------------------------------------------------------------------*/
unsigned char uartRead(unsigned char blocking)
{
    if (blocking == UART_WAIT_RECEIVE) {
        /* wait for data register to be full */
        while ( !(UCSRA & (1<<RXC)) );
    } else if (!(UCSRA & (1 << RXC))) {
        /* data register empty, return the error code */
        return UART_NO_RCV;
    }
    /* data register not empty, return the data */
    return UDR;
}

/**
 * @brief This function returns 1 when length bytes have already been received.
 * In this cae it fills buffer with length bytes ('buffer' must be an array of
 * at least length bytes!)
 */
unsigned char uartReadNBytes(unsigned char* buffer,
						     unsigned char length)
{
	unsigned char i;
	for(i=0; i < length; ++i) {
		while ( !(UCSRA & (1<<RXC)) );
		buffer[i] = UDR;
	}
	return 1;
}

#endif

#ifdef UART_TX_FIFO
static struct UartTxBuffer
{
	volatile unsigned char data[UART_TX_BUFFER_SIZE];
	         unsigned char readIndex;
	volatile unsigned char writeIndex;
	volatile unsigned char dataLength;
} txBuffer; 

/* ------------------------------------------------------------------------*/
/* uartFlushTxBuffer                                                       */
/* ------------------------------------------------------------------------*/
static void uartFlushTxBuffer()
{
	if (txBuffer.dataLength > 0) {
		unsigned char data2 = txBuffer.data[txBuffer.readIndex];
		if (++txBuffer.readIndex >= UART_TX_BUFFER_SIZE) {
			txBuffer.readIndex = 0;
		}
		--txBuffer.dataLength;
		UDR = data2;
	}
}

/* ------------------------------------------------------------------------*/
/* uartWrite                                                               */
/* ------------------------------------------------------------------------*/
void uartWrite(unsigned char data)
{
	// wait until write buffer is not full
	while(txBuffer.dataLength >= UART_TX_BUFFER_SIZE) nop();

	// add data to the buffer
	txBuffer.data[txBuffer.writeIndex] = data;
	if (++txBuffer.writeIndex >= UART_TX_BUFFER_SIZE) {
		txBuffer.writeIndex = 0;
	}
	++txBuffer.dataLength;

	// write data on the serial port if possible
	if ( ( UCSRA & (1<<UDRE)) ) {
		uartFlushTxBuffer();
	}
}

/* ------------------------------------------------------------------------ */
/* ISR(SIG_UART_DATA)                                                       */
/* ------------------------------------------------------------------------ */
/* Interrupt when write buffer is empty                                     */
/* ------------------------------------------------------------------------ */
ISR(SIG_UART_DATA)  
{
	// when the UDR register is empty, send a new data on the serial port
	uartFlushTxBuffer();
}

#else // UART_TX_FIFO
/* ------------------------------------------------------------------------*/
/* uartWrite                                                               */
/* ------------------------------------------------------------------------*/
void uartWrite(unsigned char data)
{
    /* Wait for empty transmit buffer */
    while ( !( UCSRA & (1<<UDRE)) );
    /* Put data into buffer, sends the data */
    UDR = data;
}
#endif

/* ------------------------------------------------------------------------*/
/* uartInit                                                                */
/* ------------------------------------------------------------------------*/
void uartInit(unsigned char baud)
{

#ifdef UART_RX_FIFO
	rxBuffer.readIndex=0;
	rxBuffer.writeIndex=0;
	rxBuffer.dataLength=0;
#endif
#ifdef UART_TX_FIFO
	txBuffer.readIndex=0;
	txBuffer.writeIndex=0;
	txBuffer.dataLength=0;
#endif

    //UBRR0H = 0;
    UBRRL = (unsigned char)baud;
    /* Enable receiver and transmitter + receive interrupt*/
    UCSRB = (1<<RXEN)| (1<<TXEN)
#ifdef UART_RX_FIFO
			| (1<<RXCIE)
#endif
#ifdef UART_TX_FIFO
			| (1<<UDRIE)
#endif
		;
    // 8bit, 1 stop bit, no parity
}

/* ------------------------------------------------------------------------*/
/* uartRead8                                                               */
/* ------------------------------------------------------------------------*/
unsigned char uartRead8()
{
    return uartRead(UART_WAIT_RECEIVE);
}

/* ------------------------------------------------------------------------*/
/* uartRead16                                                              */
/* ------------------------------------------------------------------------*/
uint16_t uartRead16()
{
    uint16_t data;
    data = uartRead8();
    data += (uartRead8() << 8);
    return data;
}

/* ------------------------------------------------------------------------*/
/* uartWrite16                                                             */
/* ------------------------------------------------------------------------*/
void uartWrite16(uint16_t data)
{
    uartWrite(data & 0xFF);
    uartWrite((data >> 8) & 0xFF);
}

/* ------------------------------------------------------------------------*/
/* uartWrite32                                                             */
/* ------------------------------------------------------------------------*/
void uartWrite32(uint32_t data)
{
    uartWrite(data & 0xFF);
    uartWrite((data >> 8) & 0xFF);
    uartWrite((data >> 16) & 0xFF);
    uartWrite((data >> 24) & 0xFF);
}

#ifdef UART_STRING_FUNCTIONS
/* ------------------------------------------------------------------------*/
/* uartWriteString                                                         */
/* ------------------------------------------------------------------------*/
void uartWriteString(char* s)
{
    while (*s) {
        if (*s == '\n') {
            uartWrite(0x0d);
            uartWrite(0x0a);
        } else {
            uartWrite(*s);
        }
        s++;
    }
}

static void uartWriteLetter(unsigned char b)
{
    if (b < 0x0A) {
        uartWrite('0'+b);
    } else {
        uartWrite('A'+(b- 10 ));
    }
}
void uartWriteHexToString(unsigned char data)
{
    uartWrite('0');
    uartWrite('x');
    uartWriteLetter((data>>4)&0x0F);
    uartWriteLetter(data&0x0F);
    uartWrite(' ');
}
void uartWriteShortToString(unsigned short data)
{
    uartWrite('0');
    uartWrite('x');
    uartWriteLetter((data>>12)&0x0F);
    uartWriteLetter((data>>8)&0x0F);
    uartWriteLetter((data>>4)&0x0F);
    uartWriteLetter(data&0x0F);
    uartWrite(' ');
}

/* ------------------------------------------------------------------------*/
/* uartWriteBuffer                                                         */
/* ------------------------------------------------------------------------*/
void uartWriteBuffer(unsigned char* data, 
                     unsigned short length)
{
    unsigned short i=0;
    for(i=0; i<length; ++i) {
        uartWrite(data[i]);
    }
}

#endif

