/**
 * @file       uart.h
 * @brief      serial port communication for AVR (ATmega32)
 * 
 * @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
 * @date       2009/04/26
 */

#ifndef __UART_H__
#define __UART_H__

#define UART_RX_FIFO
#define UART_RX_BUFFER_SIZE 150

//#define UART_TX_FIFO
#define UART_TX_BUFFER_SIZE 48

#if UART_RX_BUFFER_SIZE >= 256
#error UART_RX_BUFFER_SIZE must be an unsigned char
#endif

#if UART_TX_BUFFER_SIZE >= 256
#error UART_TX_BUFFER_SIZE must be an unsigned char
#endif

#define UART_STRING_FUNCTIONS

#include <inttypes.h>

/**
 * @brief Byte returned by uartReceive when there is no byte received on serial port
 * @see uartReceive
 */
#define UART_NO_RCV             0x00
/**
 * @brief uartReceive waits for a byte to be received on serial port before returning
 * @see uartReceive
 */
#define UART_WAIT_RECEIVE       0x01
/**
 * @brief uartReceive does not wait for a byte to be received on serial port before returning
 * @see uartReceive
 */
#define UART_DONT_WAIT_RECEIVE  0x00



/*************** Baud rates defined for a 8MHz cristal ***************/
/**
 * @brief Baud rate value 9600, to be used with uartInit
 */
#define UART_BAUD_RATE_2400    207
#define UART_BAUD_RATE_4800    103
#define UART_BAUD_RATE_9600	    51
#define UART_BAUD_RATE_19200	25  // 0.2%

#define UART_BAUD_RATE_38400	12  // 0.2%
#define UART_BAUD_RATE_250000	 1  // 0%

/**
 * @brief This function initialize serial port
 * @param baud rate speed for serial port,
 * @see   BAUD_RATE_9600... 
 */
void uartInit(unsigned char baud);

/**
 * @brief This function sends one byte on serial port
 */
void uartWrite(unsigned char data);
#define uartWrite8 uartWrite
/**
 * @brief This function sends 2 bytes on serial port
 */
void uartWrite16(uint16_t data);

/**
 * @brief This function sends 4 bytes on serial port
 */
void uartWrite32(uint32_t data);

/**
 * @brief This function gets one byte from serial port
 * @param blocking if set to UART_DONT_WAIT_RECEIVE, if there is no byte 
 *        waiting on serial port, the function returns immediatly the value 
 *        UART_NO_RCV\n
 *        If set to UART_WAIT_RECEIVE, the function waits until a byte
 *        arrives on serial port and it returns it
 *        
 * @see UART_WAIT_RECEIVE      
 * @see UART_DONT_WAIT_RECEIVE
 * @see UART_NO_RCV
 */
unsigned char uartRead(unsigned char blocking);

/**
 * @brief This function waits for receiving 8 bits on the serial port. 
 * Return the byte when it is received
 */
unsigned char uartRead8();
/**
 * @brief This function waits for receiving 16 bits on the serial port. 
 * Return the 2 bytes when it is received
 */
 uint16_t uartRead16();

 /**
 * @brief This function returns 1 when length bytes have already been received.
 * In this cae it fills buffer with length bytes ('buffer' must be an array of
 * at least length bytes!)
 */
unsigned char uartReadNBytes(unsigned char* buffer,
						     unsigned char length);

#ifdef UART_STRING_FUNCTIONS
/**
 * @brief This function sends a string (a text char*) on serial port
 */
 void uartWriteString(char* s);
 
 /**
 * @brief This function 
 */
 void uartWriteHexToString(unsigned char data);

/**
 * @brief This function 
 */
void uartWriteShortToString(unsigned short data);

/**
 * @brief This function send length bytes on serial port
 */
void uartWriteBuffer(unsigned char* data, unsigned short length);
#endif

#endif /* __UART_H__ */
